from .rejex_static import Static

class Rejex:
    def __init__(self, start_flag:bool=True, end_flag:bool=False) -> None:
        """ Initialize Rejex object.

        Args:
            start_flag (bool, optional): Matches expression only at the start. Defaults to True.
            end_flag (bool, optional): Matches expression only at the end(Doesn't work as intended). Defaults to False.
        """
        self.string:str = ""

        self._start_flag:bool = start_flag
        self._end_flag:bool = end_flag

        if not start_flag:
            self.string += ".*"

    @property
    def get_start_flag(self) -> bool:
        return self._start_flag

    @property
    def get_end_flag(self) -> bool:
        return self._end_flag

    def compile(self) -> str:
        if self._end_flag:
            self.string += "$"

        return self.string

    def literal(self, literal:str) -> 'Rejex':
        """ Literal is a string that will be matched exactly.

        Args:
            literal (str): Literal string to be matched.

        Returns:
            Rejex: Rejex object.
        """

        self.string += Static.literal(literal)

        return self

    def any_char_except_newline(self) -> 'Rejex':
        """ Any character except newline.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.any_char_except_newline()

        return self

    def any_number(self) -> 'Rejex':
        """ Match any number character. Similar to '[0-9]'.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.any_number()

        return self

    def any_non_number(self) -> 'Rejex':
        """ Match any non number character. Similar to '[0-9]'.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.any_non_number()

        return self

    def any_word_character(self) -> 'Rejex':
        """ Match any word character. Similar to '[a-zA-Z_0-9]'.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.any_word_character()

        return self

    def any_non_word_character(self) -> 'Rejex':
        """ Match any non word character. Similar to '[^a-zA-Z_0-9]'.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.any_non_word_character()

        return self

    def any_whitespace_character(self) -> 'Rejex':
        """ Match any whitespace character. Similar to '[ \\f\\n\\r\\t\\v]'.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.any_whitespace_character()

        return self

    def any_non_whitespace_character(self) -> 'Rejex':
        """ Match any non whitespace character. Similar to '[^ \\f\\n\\r\\t\\v]'.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.any_non_whitespace_character()

        return self

    def alternative(self, *alts, fast=False) -> 'Rejex':
        """ If a string matches any one in the alts list then it will match.

        Args:
            alts (list): The list of alt strings/chars.
            fast (bool): Doesn't check alts and compiles to string using the fastest universal method(Might become unreadable). Defaults to False

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.alternative(*alts, fast=fast)

        return self

    def except_expr(self, expr:str) -> 'Rejex':
        """ Is a NOT operation on the expression.
        This doesn't work with "start_flag=False"

        Args:
            expr (str): The expression to be inversed

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.except_expr(expr)

        return self

    def n_number_times(self, n:int, expr:str) -> 'Rejex':
        """ Match 'expr' expression set exactly 'n' number of times.

        Args:
            n (int): The number of times the previous set will be matched/searched for.
            expr (str): The expression set to be matched exactly 'n' number of times.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.n_number_times(n, expr)

        return self

    def zero_or_more(self, expr:str) -> 'Rejex':
        """ The 'expr' to be matched zero or more times.

        Args:
            expr (str): The 'exp' to be matched zero or more times.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.zero_or_more(expr)

        return self

    def one_or_more(self, expr:str) -> 'Rejex':
        """ The 'expr' to be matched one or more times.

        Args:
            expr (str): The 'exp' to be matched one or more times.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.one_or_more(expr)

        return self

    def zero_or_one(self, expr:str) -> 'Rejex':
        """ The 'expr' to be matched zero or one time.

        Args:
            expr (str): The 'exp' to be matched zero or one time.

        Returns:
            Rejex: Rejex object.
        """
        self.string += Static.zero_or_one(expr)

        return self

    def range(self, start:str, end:str) -> 'Rejex':
        """ Match a range of characters.

        Args:
            start (str): The start of the range.
            end (str): The end of the range.

        Returns:
            str: Rejex string.
        """
        self.string += Static.range(start, end)

        return self


"""Rejex is a tool used to build regex expressions for python."""

from .rejex_class import Rejex
from .rejex_static import *

def test_regex(regex_pattern:str, string:str, flag:int=0) -> bool:
    matched = bool(match(regex_pattern, string, flag))
    print(f"-------------------------------\nregex_pattern='{regex_pattern}'\nstring='{string}'\nmatch={matched}\n-------------------------------")

class CommonRejex:
    phone_number: str = r"(\+(\d)+)? ?(\()?(\d){3}(\))? ?(\d){3}((-| ))?(\d){4}"
    """ 
    (
        Rejex()
            .zero_or_one(
                Rejex()
                    .literal("+")
                    .one_or_more(
                        Static.any_number()
                    )
                    .compile()
            )
            .zero_or_one(Static.literal(" "))
            .zero_or_one(Static.literal("("))
            .n_number_times(3, Static.any_number())
            .zero_or_one(Static.literal(")"))
            .zero_or_one(Static.literal(" "))
            .n_number_times(3, Static.any_number())
            .zero_or_one(Static.alternative(Static.literal("-"), Static.literal(" ")))
            .n_number_times(4, Static.any_number())
        .compile()
    )
    ...
    """

    short_date:str = r"(\d)?\d(/|-)(\d)?\d(/|-)((\d){2})?(\d){2}"
    """
    (
        Rejex()
            .zero_or_one(Static.any_number())
            .any_number()
            .alternative(Static.literal("/"), Static.literal("-"))
            .zero_or_one(Static.any_number())
            .any_number()
            .alternative(Static.literal("/"), Static.literal("-"))
            .zero_or_one(
                Static.n_number_times(2, Static.any_number())
            )
            .n_number_times(2, Static.any_number())
        .compile()
    )"""

    basic_time:str = r"[0-9]?[0-9]:[0-9][0-9](:[0-9][0-9])?"


from re import *

class Static:

    @staticmethod
    def literal(literal:str) -> str:
        """ Literal is a string that will be matched exactly.

        Args:
            literal (str): Literal string to be matched.

        Returns:
            str: Rejex string.
        """
        if not isinstance(literal, str):
            try:
                literal = str(literal)
            except Exception:
                raise ValueError(f"Literal '{literal}' must be a string or should be convertable to string.")

        to_be_escaped = "^$.|()*+?{}[]\\"
        new_literal = ""
        for i in literal:
            if i not in to_be_escaped:
                new_literal += i
            else:
                new_literal += "\\" + i

        return new_literal

    @staticmethod
    def any_char_except_newline() -> str:
        """ Any character except newline. If DOTALL flag is selected, newlines are also matched

        Returns:
            str: Rejex string.
        """
        return "."

    @staticmethod
    def any_number() -> str:
        """ Match any number character. Similar to '[0-9]'.

        Returns:
            str: Rejex string.
        """
        return "\\d"

    @staticmethod
    def any_non_number() -> str:
        """ Match any non number character. Similar to '[^0-9]'.

        Returns:
            str: Rejex string.
        """
        return "\\D"

    @staticmethod
    def any_word_character() -> str:
        """ Match any word character. Similar to '[a-zA-Z_0-9]'.

        Returns:
            str: Regex string.
        """
        return "\\w"

    @staticmethod
    def any_non_word_character() -> str:
        """ Match any non word character. Similar to '[^a-zA-Z_0-9]'.

        Returns:
            str: Regex string.
        """
        return "\\W"

    @staticmethod
    def any_whitespace_character() -> str:
        """ Match any whitespace character. Similar to '[ \\f\\n\\r\\t\\v]'.

        Returns:
            str: Regex string.
        """
        return "\\s"

    @staticmethod
    def any_non_whitespace_character() -> str:
        """ Match any non whitespace character. Similar to '[^ \\f\\n\\r\\t\\v]'.

        Returns:
            str: Regex string.
        """
        return "\\S"

    @staticmethod
    def alternative(*alts, fast=False) -> str:
        """ If a string matches any one in the alts list then it will match.

        Args:
            alts (list): The list of alt strings/chars.

        Returns:
            str: Rejex string.
        """
        alts = [*alts]
        try:
            new_alts = [i if isinstance(i, str) else str(i) for i in alts]
        except Exception:
            raise ValueError(f"All values in alts '{alts}' must be a string or should be convertable to string.")

        if fast:
            return f"({'|'.join(new_alts)})"

        if all(len(i)==1 and i != "-" for i in new_alts):
            #if all of them are single chars, use [abc] notation
            return f"[{''.join(new_alts)}]"
        elif all(bool(match(r"^\[.-.\]$", i)) for i in new_alts):
            return f"[{''.join(i[1:-1] for i in new_alts)}]"
        else:
            #if any of them are more than one char, use a|b|c notation
            return f"({'|'.join(new_alts)})"

    @staticmethod
    def except_expr(expr:str) -> str:
        """ Is a NOT operation on the expression.
        This doesn't work with "start_flag=False"

        Args:
            expr (str): The expression to be inversed

        Returns:
            str: Rejex string.
        """
        if not isinstance(expr, str):
            try:
                expr = str(expr)
            except ValueError:
                raise ValueError(f"Expression '{expr}' must be a string or should be convertable to string.")

        if expr[0] == "[" and expr[-1] == "]":
            return f"[^{expr[1:-1]}]"

        return f"[^{expr}]"

    @staticmethod
    def n_number_times(n:int, expr:str) -> str:
        """ Match 'expr' expression set exactly 'n' number of times.

        Args:
            n (int): The number of times the previous set will be matched/searched for.
            expr (str): The expression set to be matched exactly 'n' number of times.

        Returns:
            str: Rejex string.
        """
        assert isinstance(n, int), "n must be an integer."
        
        if not isinstance(expr, str):
            try:
                expr = str(expr)
            except ValueError:
                raise ValueError(f"Expression '{expr}' must be a string or should be convertable to string.")

        if len(expr) == 1:
            return f"{expr}{{{n}}}"
        return f"({expr}){{{n}}}"

    @staticmethod
    def zero_or_more(expr:str) -> str:
        """ The 'expr' to be matched zero or more times.

        Args:
            expr (str): The 'exp' to be matched zero or more times.

        Returns:
            str: Rejex string.
        """
        if not isinstance(expr, str):
            try:
                expr = str(expr)
            except ValueError:
                raise ValueError(f"Expression '{expr}' must be a string or should be convertable to string.")

        if len(expr) == 1:
            return f"{expr}*"
        return f"({expr})*"

    @staticmethod
    def one_or_more(expr:str) -> str:
        """ The 'expr' to be matched one or more times.

        Args:
            expr (str): The 'exp' to be matched one or more times.

        Returns:
            str: Rejex string.
        """
        if not isinstance(expr, str):
            try:
                expr = str(expr)
            except ValueError:
                raise ValueError(f"Expression '{expr}' must be a string or should be convertable to string.")

        if len(expr) == 1:
            return f"{expr}+"
        return f"({expr})+"

    @staticmethod
    def zero_or_one(expr:str) -> str:
        """ The 'expr' to be matched zero or one time.

        Args:
            expr (str): The 'exp' to be matched zero or one time.

        Returns:
            str: Rejex string.
        """
        if not isinstance(expr, str):
            try:
                expr = str(expr)
            except ValueError:
                raise ValueError(f"Expression '{expr}' must be a string or should be convertable to string.")

        if len(expr) == 1:
            return f"{expr}?"
        return f"({expr})?"

    @staticmethod
    def range(start:str, end:str) -> str:
        """ Match a range of characters.

        Args:
            start (str): The start of the range.
            end (str): The end of the range.

        Returns:
            str: Rejex string.
        """
        if not isinstance(start, str):
            try:
                start = str(start)
            except ValueError:
                raise ValueError(f"Start '{start}' must be a string or should be convertable to string.")

        if not isinstance(end, str):
            try:
                end = str(end)
            except ValueError:
                raise ValueError(f"End '{end}' must be a string or should be convertable to string.")

        return f"[{start}-{end}]"